@@ -183,6 +183,14 @@ static void gtk_theming_engine_render_icon (GtkThemingEngine *engine,
 					    GdkPixbuf *pixbuf,
                                             gdouble x,
                                             gdouble y);
+static void render_background_internal (GtkThemingEngine *engine,
+					cairo_t          *cr,
+					gdouble           x,
+					gdouble           y,
+					gdouble           width,
+					gdouble           height,
+					GtkJunctionSides  junction,
+					cairo_pattern_t  *optional_background);
 
 G_DEFINE_TYPE (GtkThemingEngine, gtk_theming_engine, G_TYPE_OBJECT)
 
@@ -450,6 +458,25 @@ gtk_theming_engine_get_valist (GtkThemingEngine *engine,
   gtk_style_context_get_valist (priv->context, state, args);
 }
 
+void
+_gtk_theming_engine_get (GtkThemingEngine *engine,
+			 GtkStateFlags     state,
+			 GtkStylePropertyContext *property_context,
+			 ...)
+{
+  GtkThemingEnginePrivate *priv;
+  va_list args;
+
+  g_return_if_fail (GTK_IS_THEMING_ENGINE (engine));
+
+  priv = engine->priv;
+
+  va_start (args, property_context);
+  _gtk_style_context_get_valist (priv->context, state, property_context, args);
+  va_end (args);
+}
+
+
 /**
  * gtk_theming_engine_get:
  * @engine: a #GtkThemingEngine
@@ -1066,6 +1093,8 @@ gtk_theming_engine_render_check (GtkThemingEngine *engine,
   GtkBorderStyle border_style;
   GtkBorder border;
   gint border_width;
+  GtkStylePropertyContext context;
+  cairo_pattern_t *pattern;
 
   flags = gtk_theming_engine_get_state (engine);
   cairo_save (cr);
@@ -1074,9 +1103,22 @@ gtk_theming_engine_render_check (GtkThemingEngine *engine,
   gtk_theming_engine_get_background_color (engine, flags, &bg_color);
   gtk_theming_engine_get_border (engine, flags, &border);
 
-  gtk_theming_engine_get (engine, flags,
-                          "border-style", &border_style,
-                          NULL);
+  context.width = width;
+  context.height = height;
+
+  _gtk_theming_engine_get (engine, flags, &context,
+			   "background-image", &pattern,
+			   "border-style", &border_style,
+			   NULL);
+
+  if (pattern != NULL)
+    {
+      render_background_internal (engine, cr, x, y, width, height, 
+				  gtk_theming_engine_get_junction_sides (engine), pattern);
+      cairo_restore (cr);
+      cairo_pattern_destroy (pattern);
+      return;
+    }
 
   border_width = MIN (MIN (border.top, border.bottom),
                       MIN (border.left, border.right));
@@ -1186,6 +1228,8 @@ gtk_theming_engine_render_option (GtkThemingEngine *engine,
   gint exterior_size, interior_size, pad, thickness, border_width;
   GtkBorderStyle border_style;
   GtkBorder border;
+  GtkStylePropertyContext context;
+  cairo_pattern_t *pattern;
 
   flags = gtk_theming_engine_get_state (engine);
 
@@ -1195,9 +1239,22 @@ gtk_theming_engine_render_option (GtkThemingEngine *engine,
   gtk_theming_engine_get_background_color (engine, flags, &bg_color);
   gtk_theming_engine_get_border (engine, flags, &border);
 
-  gtk_theming_engine_get (engine, flags,
-                          "border-style", &border_style,
-                          NULL);
+  context.width = width;
+  context.height = height;
+
+  _gtk_theming_engine_get (engine, flags, &context,
+			   "background-image", &pattern,
+			   "border-style", &border_style,
+			   NULL);
+
+  if (pattern != NULL)
+    {
+      render_background_internal (engine, cr, x, y, width, height, 
+				  gtk_theming_engine_get_junction_sides (engine), pattern);
+      cairo_restore (cr);
+      cairo_pattern_destroy (pattern);
+      return;
+    }
 
   exterior_size = MIN (width, height);
   border_width = MIN (MIN (border.top, border.bottom),
@@ -1370,6 +1427,9 @@ color_shade (const GdkRGBA *color,
   gtk_symbolic_color_unref (shade);
 }
 
+/* optional_background is the background-image, in case it was
+   already availible we avoid getting it (and thus possibly
+   rendering it) twice */
 static void
 render_background_internal (GtkThemingEngine *engine,
                             cairo_t          *cr,
@@ -1377,25 +1437,38 @@ render_background_internal (GtkThemingEngine *engine,
                             gdouble           y,
                             gdouble           width,
                             gdouble           height,
-                            GtkJunctionSides  junction)
+                            GtkJunctionSides  junction,
+			    cairo_pattern_t  *optional_background)
 {
   GdkRGBA bg_color;
   cairo_pattern_t *pattern;
+  GtkCssBackgroundRepeat *repeat;
   GtkStateFlags flags;
   gboolean running;
   gdouble progress;
   GtkRoundedBox border_box;
   GtkShadow *box_shadow;
   GtkBorder border;
+  GtkStylePropertyContext context;
 
   flags = gtk_theming_engine_get_state (engine);
 
   gtk_theming_engine_get_background_color (engine, flags, &bg_color);
 
-  gtk_theming_engine_get (engine, flags,
-                          "background-image", &pattern,
-                          "box-shadow", &box_shadow,
-                          NULL);
+  context.width = width;
+  context.height = height;
+
+  if (optional_background)
+    pattern = cairo_pattern_reference (optional_background);
+  else
+    _gtk_theming_engine_get (engine, flags, &context, 
+			     "background-image", &pattern,
+			     NULL);
+
+  _gtk_theming_engine_get (engine, flags, &context, 
+			   "background-repeat", &repeat,
+			   "box-shadow", &box_shadow,
+			   NULL);
 
   cairo_save (cr);
   cairo_translate (cr, x, y);
@@ -1425,9 +1498,9 @@ render_background_internal (GtkThemingEngine *engine,
         other_flags = flags | GTK_STATE_FLAG_PRELIGHT;
 
       gtk_theming_engine_get_background_color (engine, other_flags, &other_bg);
-      gtk_theming_engine_get (engine, other_flags,
-                              "background-image", &other_pattern,
-                              NULL);
+      _gtk_theming_engine_get (engine, other_flags, &context,
+			       "background-image", &other_pattern,
+			       NULL);
 
       if (pattern && other_pattern)
         {
@@ -1602,9 +1675,32 @@ render_background_internal (GtkThemingEngine *engine,
 
   if (pattern)
     {
-      cairo_scale (cr, width, height);
+      cairo_surface_t *surface;
+      int scale_width, scale_height;
+
+      if (cairo_pattern_get_surface (pattern, &surface) != CAIRO_STATUS_SUCCESS)
+          surface = NULL;
+
+      if (surface && repeat &&
+          repeat->repeat != GTK_CSS_BACKGROUND_REPEAT_STYLE_NONE)
+        {
+          scale_width = cairo_image_surface_get_width (surface);
+          scale_height = cairo_image_surface_get_height (surface);
+          if (repeat->repeat == GTK_CSS_BACKGROUND_REPEAT_STYLE_REPEAT)
+            cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
+          else if (repeat->repeat == GTK_CSS_BACKGROUND_REPEAT_STYLE_NO_REPEAT)
+            cairo_pattern_set_extend (pattern, CAIRO_EXTEND_NONE);
+        }
+      else
+        {
+          cairo_pattern_set_extend (pattern, CAIRO_EXTEND_PAD);
+          scale_width = width;
+          scale_height = height;
+        }
+
+      cairo_scale (cr, scale_width, scale_height);
       cairo_set_source (cr, pattern);
-      cairo_scale (cr, 1.0 / width, 1.0 / height);
+      cairo_scale (cr, 1.0 / scale_width, 1.0 / scale_height);
     }
   else
     gdk_cairo_set_source_rgba (cr, &bg_color);
@@ -1614,6 +1710,9 @@ render_background_internal (GtkThemingEngine *engine,
   if (pattern)
     cairo_pattern_destroy (pattern);
 
+  if (repeat)
+    g_free (repeat);
+
   if (box_shadow != NULL)
     {
       _gtk_box_shadow_render (box_shadow, cr, &border_box);
@@ -1637,7 +1736,7 @@ gtk_theming_engine_render_background (GtkThemingEngine *engine,
 
   render_background_internal (engine, cr,
                               x, y, width, height,
-                              junction);
+                              junction, NULL);
 }
 
 static void
@@ -1823,15 +1922,19 @@ gtk_theming_engine_render_frame (GtkThemingEngine *engine,
   GtkJunctionSides junction;
   GtkBorderImage *border_image;
   GtkBorder border;
+  GtkStylePropertyContext context;
 
   flags = gtk_theming_engine_get_state (engine);
   junction = gtk_theming_engine_get_junction_sides (engine);
   gtk_theming_engine_get_border (engine, flags, &border);
 
-  gtk_theming_engine_get (engine, flags,
-                          "border-image", &border_image,
-                          "border-style", &border_style,
-                          NULL);
+  context.width = width;
+  context.height = height;
+
+  _gtk_theming_engine_get (engine, flags, &context,
+			   "border-image", &border_image,
+			   "border-style", &border_style,
+			   NULL);
 
   if (border_image != NULL)
     {
@@ -2341,11 +2444,11 @@ gtk_theming_engine_render_extension (GtkThemingEngine *engine,
       gap_side == GTK_POS_BOTTOM)
     render_background_internal (engine, cr,
                                 0, 0, width, height,
-                                GTK_JUNCTION_BOTTOM);
+                                GTK_JUNCTION_BOTTOM, NULL);
   else
     render_background_internal (engine, cr,
                                 0, 0, height, width,
-                                GTK_JUNCTION_BOTTOM);
+                                GTK_JUNCTION_BOTTOM, NULL);
   cairo_restore (cr);
 
   cairo_save (cr);
@@ -2411,7 +2514,7 @@ gtk_theming_engine_render_handle (GtkThemingEngine *engine,
   color_shade (&bg_color, 0.7, &darker);
   color_shade (&bg_color, 1.3, &lighter);
 
-  render_background_internal (engine, cr, x, y, width, height, sides);
+  render_background_internal (engine, cr, x, y, width, height, sides, NULL);
 
   if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_GRIP))
     {
